/* *****************************************************************************
 * core/stream.c -- libp7 easy stream usage.
 * Copyright (C) 2016-2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 *
 * These functions are the ~equivalents of `fread`/`fwrite` for libp7 streams,
 * with communication settings function. It is highly advised to use them
 * instead of using the streams reading directly.
 * ************************************************************************** */
#include <libp7/internals.h>
#include <string.h>

/* ************************************************************************** */
/*  Main stream callbacks                                                     */
/* ************************************************************************** */
/**
 *	p7_read:
 *	Receive raw data via a P7 Stream.
 *
 *	@arg	stream		the P7 stream.
 *	@arg	dest		the data destination.
 *	@arg	size		the data size.
 *	@return				the error code (0 if ok).
 */

int p7_read(p7_stream_t *stream, void *dest, size_t size)
{
	int err;
	if (stream->p7_stream_flags & p7_streamflag_scsi || !stream->p7_stream_read)
		return (p7_error_noread);
	if (!size) return (0);

	/* filter start things
	 * FIXME: maybe they're useful?
	 * FIXME: this won't help the legacy protocol lol */
	unsigned char byte;
	do {
		err = (*stream->p7_stream_read)(stream->p7_stream_cookie, &byte, 1);
		if (err) return (err);
	} while (byte == 0x10 || byte == 0x16);

	/* return real result */
	unsigned char *d = (void*)dest; *d++ = byte; size--;
	err = (*stream->p7_stream_read)(stream->p7_stream_cookie, d, size);
	return (err);
}

/**
 *	p7_write:
 *	Send raw data via a P7 Stream.
 *
 *	@arg	stream		the P7 stream.
 *	@arg	data		the data.
 *	@arg	size		the data size.
 *	@return				the error code (0 if ok).
 */

int p7_write(p7_stream_t *stream, const void *data, size_t size)
{
	if (stream->p7_stream_flags & p7_streamflag_scsi
	 || !stream->p7_stream_write)
		return (p7_error_nowrite);
	return ((*stream->p7_stream_write)(stream->p7_stream_cookie,
		(const unsigned char*)data, size));
}
/* ************************************************************************** */
/*  Set stream communication settings and timeouts                            */
/* ************************************************************************** */
/**
 *	p7_setcomm:
 *	Change the settings of a P7 Stream.
 *
 *	@arg	stream		the P7 stream.
 *	@arg	settings	the settings to set.
 *	@return				the error code (0 if ok).
 */

int p7_setcomm(p7_stream_t *stream, const p7_streamsettings_t *settings)
{
	if (stream->p7_stream_flags & p7_streamflag_usb) {
		logr_info("Stream is an USB device!");
		return (0);
	}
	if (!stream->p7_stream_setcomm) {
		logr_info("No communication manager for this stream.");
		return (0);
	}

	/* get the default settings if necessary */
	p7_streamsettings_t localsettings;
	if (!settings) {
		settings = &localsettings;
		p7_initcomm(&localsettings);
	}

	/* describe the settings */
#if LOGLEVEL <= ll_info
	const char *es[] = {"disabled", "enabled", "handshake", "disabled"};

	logr_info("Setting serial settings: %u%c%c",
		settings->p7_streamsettings_speed,
		~settings->p7_streamsettings_flags & P7_PARENB ? 'N'
		: settings->p7_streamsettings_flags & P7_PARODD ? 'O' : 'E',
		settings->p7_streamsettings_flags & P7_TWOSTOPBITS ? '2' : '1');
	logr_info("DTR mode: %s, RTS mode: %s",
		es[P7_DTRVAL(settings->p7_streamsettings_flags)],
		es[P7_RTSVAL(settings->p7_streamsettings_flags)]);

	if (settings->p7_streamsettings_flags & P7_XONMASK)
		logr_info("XON is enabled (0x%02X)",
			settings->p7_streamsettings_cc[P7_XON]);
	else logr_info("XON is disabled");
	if (settings->p7_streamsettings_flags & P7_XOFFMASK)
		logr_info("XOFF is enabled (0x%02X)",
			settings->p7_streamsettings_cc[P7_XOFF]);
	else logr_info("XOFF is disabled");
#endif

	/* then make */
	return ((*stream->p7_stream_setcomm)(stream->p7_stream_cookie, settings));
}

/**
 *	p7_settm:
 *	Change the timeouts of a P7 Stream.
 *
 *	@arg	stream		the P7 stream.
 *	@arg	settings	the settings to set.
 *	@return				the error code (0 if ok).
 */

int p7_settm(p7_stream_t *stream, const p7_streamtimeouts_t *timeouts)
{
	if (!stream->p7_stream_settm) return (0);
	return ((*stream->p7_stream_settm)(stream->p7_stream_cookie, timeouts));
}
