/* *****************************************************************************
 * utils/settings.c -- stream settings-related utilities.
 * Copyright (C) 2016-2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libp7.
 * libp7 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libp7 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libp7; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libp7/internals.h>
#include <string.h>

/* ************************************************************************** */
/*  Initialize communication settings                                         */
/* ************************************************************************** */
/* default settings */
static p7_streamsettings_t default_settings = {
	.p7_streamsettings_flags =
		P7_TWOSTOPBITS | P7_DTRCTL_ENABLE | P7_RTSCTL_ENABLE,
	.p7_streamsettings_speed = P7_B9600,
};

/**
 *	p7_initcomm:
 *	Initialize a communication settings structure.
 *
 *	@arg	settings	the settings to initialize.
 */

void p7_initcomm(p7_streamsettings_t *settings)
{
	memcpy(settings, &default_settings, sizeof(p7_streamsettings_t));
}
/* ************************************************************************** */
/*  Make settings from a string                                               */
/* ************************************************************************** */
/**
 *	p7_makesettings:
 *	Make settings out of a string.
 *
 *	@arg	settings	the settings to make.
 *	@arg	s			the string.
 *	@return				the error code (0 if ok).
 */

int p7_makesettings(p7_streamsettings_t *settings, const char *s)
{
	/* extract data from the string */
	unsigned int speed; char par; int stop;
	if (sscanf(s, "%u%c%d", &speed, &par, &stop) < 3
	 || (par != 'N' && par != 'E' && par != 'O')
	 || (stop != 1 && stop != 2)) {
		logr_error("Invalid format!");
		return (p7_error_invalid);
	}

	/* get the speed */
	switch (speed) {
	case 1200:   speed = P7_B1200;   break;
	case 2400:   speed = P7_B2400;   break;
	case 4800:   speed = P7_B4800;   break;
	case 9600:   speed = P7_B9600;   break;
	case 19200:  speed = P7_B19200;  break;
	case 38400:  speed = P7_B38400;  break;
	case 57600:  speed = P7_B57600;  break;
	case 115200: speed = P7_B115200; break;
	default:
		logr_error("%u isn't a valid speed!", speed);
		return (p7_error_invalid);
	}

	/* set the settings */
	p7_initcomm(settings);
	settings->p7_streamsettings_speed = speed;
	settings->p7_streamsettings_flags &= ~(P7_TWOSTOPBITS | P7_PARMASK);
	settings->p7_streamsettings_flags |= --stop ? P7_TWOSTOPBITS : P7_ONESTOPBIT;
	settings->p7_streamsettings_flags |= par == 'N' ? P7_PARDIS :
		par == 'E' ? P7_PARENB | P7_PAREVEN : P7_PARENB | P7_PARODD;
	return (0);
}
